use std::{collections::HashMap, process::exit};

use clap::error::Result;
use codespan_reporting::{
    files::SimpleFiles,
    term::{
        self,
        termcolor::{ColorChoice, StandardStream},
    },
};

use crate::{
    builtins::initialise_globals,
    error::{ErrorKind, Errors},
    syntax::{check::check, parse_syntax, PipelineElement},
    typed::into_typed_repr,
};

pub struct EvalConfig {
    debug_raw_toks: bool,
    debug_print_typed_repr: bool,
}

impl EvalConfig {
    pub fn new(debug_raw_toks: bool, debug_print_typed_repr: bool) -> Self {
        Self {
            debug_raw_toks,
            debug_print_typed_repr,
        }
    }
}

// this is also bad
// need a better architecture for this
pub struct Evaluator<'a> {
    curr_phase: EvalPhase,
    files: SimpleFiles<&'a str, String>,
    errors: HashMap<usize, Vec<Errors>>,
    cfg: EvalConfig,
}

impl<'a> Evaluator<'a> {
    pub fn init(cfg: EvalConfig) -> Self {
        Self {
            curr_phase: EvalPhase::Lex,
            files: SimpleFiles::new(),
            errors: HashMap::new(),
            cfg,
        }
    }

    pub fn run(&mut self, input: String, name: Option<&'a str>) {
        let fid = self.files.add(name.unwrap_or("input"), input.clone());

        let syntax = parse_syntax(&input);
        if self.cfg.debug_raw_toks {
            println!("Raw tokens: {syntax:#?}");
        }

        match syntax {
            Ok(syntax) => self.curr_phase = EvalPhase::Check(fid, syntax),
            Err(errs) => {
                self.errors.insert(
                    fid,
                    vec![Errors {
                        kind: ErrorKind::InvalidToken,
                        locs: errs,
                    }],
                );
                self.curr_phase = EvalPhase::Failed;
            }
        };
    }

    pub fn next(mut self) {
        match self.curr_phase {
            EvalPhase::Lex => {
                todo!()
            }
            EvalPhase::Check(file_id, syntax) => {
                let r = check(&syntax);

                if let Err(errs) = r {
                    self.errors.insert(file_id, errs);
                    self.curr_phase = EvalPhase::Failed;
                } else {
                    self.curr_phase = EvalPhase::BareTyped(file_id, syntax.clone());
                }
            }
            EvalPhase::BareTyped(file_id, syntax) => {
                let ns = initialise_globals();
                let r = into_typed_repr(&ns, syntax);

                match r {
                    Ok(typed) => {
                        if self.cfg.debug_print_typed_repr {
                            println!("Typed repr: {typed:#?}");
                        }

                        todo!()
                    }
                    Err(errs) => {
                        self.errors.insert(file_id, vec![errs]);
                        self.curr_phase = EvalPhase::Failed;
                    }
                }
            }
            EvalPhase::Failed => self.error_out().expect("unable to print errors"),
        }
        self.next();
    }

    pub fn error_out(self) -> Result<!, codespan_reporting::files::Error> {
        let Evaluator { files, errors, .. } = self;

        let writer = StandardStream::stderr(ColorChoice::Always);
        let config = term::Config::default();

        for (file_id, errors) in &errors {
            let writer = &mut writer.lock();
            for error in errors {
                term::emit(writer, &config, &files, &error.into_diag(*file_id, &files))?;
            }
        }

        exit(1)
    }
}

enum EvalPhase {
    Lex,
    Check(usize, Vec<PipelineElement>),
    BareTyped(usize, Vec<PipelineElement>),
    Failed,
}
